/*
Bounding volumes
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "vector.h"
#include "surface.h"
#include "object.h"
#include "bvol.h"

#include "sphere.h"
#include "disc.h"
#include "triangle.h"



inline bvol_p bvol_new()
{
 bvol_p r = malloc(sizeof(bvol_s));
 r->child1 = NULL;
 r->child2 = NULL;
 r->data = NULL;
 return r;
}






inline void ray_extra_fill(ray_extra_p re, ray_p ray)
{
 re->invdir[0] = 1.0 / ray->dir[0];
 re->invdir[1] = 1.0 / ray->dir[1];
 re->invdir[2] = 1.0 / ray->dir[2];

 re->signs[0] = (re->invdir[0] < 0.0) ? 0 : 1;
 re->signs[1] = (re->invdir[1] < 0.0) ? 0 : 1;
 re->signs[2] = (re->invdir[2] < 0.0) ? 0 : 1;
}



inline vectype bbox_test_ray_ints(bbox_p bbox, ray_p ray, ray_extra_p re)
{
 vectype tmin, tmax, tymin, tymax, tzmin, tzmax; 

 tmax = (bbox->bounds[re->signs[0]][0] - ray->point[0]) * re->invdir[0];
 if(tmax < 0) return tmax; 
 tmin = (bbox->bounds[1 - re->signs[0]][0] - ray->point[0]) * re->invdir[0];

 tymax = (bbox->bounds[re->signs[1]][1] - ray->point[1]) * re->invdir[1];
 if(tymax < 0) return tymax; 
 tymin = (bbox->bounds[1 - re->signs[1]][1] - ray->point[1]) * re->invdir[1];

 if((tmin > tymax) || (tymin > tmax)) return -1;

 if(tymin > tmin) tmin = tymin;
 if(tymax < tmax) tmax = tymax;

 tzmax = (bbox->bounds[re->signs[2]][2] - ray->point[2]) * re->invdir[2];
 if(tzmax < 0) return tzmax; 
 tzmin = (bbox->bounds[1 - re->signs[2]][2] - ray->point[2]) * re->invdir[2];

 if((tmin > tzmax) || (tzmin > tmax)) return -2;
 
 if(tzmin > tmin) tmin = tzmin;
 if(tzmax < tmax) tmax = tzmax;

 return tmin<0?0.0:tmin;
}


/*
 if(ray->point[0] >= bbox->min[0] &&
    ray->point[0] <= bbox->max[0] &&
    ray->point[1] >= bbox->min[1] &&
    ray->point[1] <= bbox->max[1] &&
    ray->point[2] >= bbox->min[2] &&
    ray->point[2] <= bbox->max[2])
    {
     return 0.0;
    }
 */





/* maximum level of nested bounding boxes to search.  exceeding this will result in
bad things. */
#define MAXLEVEL 100


vectype bvol_ints_first_sub(ints_p ints, bvol_p tree, ray_p ray, obj ign)
{
 int level = 0;
 vectype d1, d2;
 bvol_p trees[MAXLEVEL * 2];
 ints_s best_ints = *ints;

 ray_extra_s re;

 trees[level] = tree;
 best_ints.dist = -1;
 ray_extra_fill(&re, ray);


/*
 if(bbox_test_ray_ints(&(tree->bbox), ray) < 0)
    {
     fprintf(stderr, "Does not intersect world!\n");
     return -103;
    }
*/

 while(level >= 0)
    {
     if(level >= MAXLEVEL * 2) 
        {
         fprintf(stderr, "Too many levels of bounding boxes - aborting search!\n");
         return -100;
        }
/*
     if(trees[level] == NULL)
        {
         fprintf(stderr, "Null tree node!\n");
         return -101;
        }
*/

     if(trees[level]->child1 == NULL || trees[level]->child2 == NULL)
        {
         /* fprintf(stderr, "Leaf node, testing with object.\n"); */
/*
         if(trees[level]->data == NULL)
            {
             fprintf(stderr, "Bounding volume does not contain an object!\n"); 
             return -102;
            }
*/
         if(trees[level]->data == ign)
            {
             /* fprintf(stderr, "Ignoring object!\n"); */
             level--;
             continue;
            }
         ints->dist = obj_test(ints->target, trees[level]->data, ray);
         if(ints->dist >= 0)
            {
             /* fprintf(stderr, "Intersected object!\n"); */
             if(ints->dist < best_ints.dist || best_ints.dist < 0)
                {
                 ints->data = trees[level]->data;
                 best_ints = *ints;
                }
            }
            else
            {
             /* fprintf(stderr, "Did not intersect object.\n"); */
            }

         level--;
         continue;
        }


     d1 = bbox_test_ray_ints(&(trees[level]->child1->bbox), ray, &re);
     d2 = bbox_test_ray_ints(&(trees[level]->child2->bbox), ray, &re);
     if(d1 >= 0 && d2 >= 0 && ((d1 < best_ints.dist && d2 < best_ints.dist) || best_ints.dist < 0) )
        {
         if(d1 - 0.2 < d2)
            {
             trees[level+1] = trees[level]->child1;
             trees[level] = trees[level]->child2;
             level++;
             continue;
            }
            else
            {
             trees[level+1] = trees[level]->child2;
             trees[level] = trees[level]->child1;
             level++;
             continue;
            }
        }
     if(d1 >= 0 && (d1 < best_ints.dist || best_ints.dist < 0))
        {
         trees[level] = trees[level]->child1;
         continue;
        }
     if(d2 >= 0 && (d2 < best_ints.dist || best_ints.dist < 0))
        {
         trees[level] = trees[level]->child2;
         continue;
        }

     /* fprintf(stderr, "Intersected with box, but neither child was suitable.\n"); */
     level--;
    }


 *ints = best_ints;
 return ints->dist;
}



inline vectype bvol_ints_first(ints_p ints, bvol_p tree, ray_p ray)
{
 ints->bits = 0x0;
 ints->dist = bvol_ints_first_sub(ints, tree, ray, NULL);
 ints->incray = *ray;
 return ints->dist;
}

inline vectype bvol_ints_first_ignore(ints_p ints, bvol_p tree, ray_p ray, obj ign)
{
 ints->bits = 0x0;
 ints->dist = bvol_ints_first_sub(ints, tree, ray, ign);
 ints->incray = *ray;
 return ints->dist;
}







/*
inline void bbox_combine(bbox_p r, bbox_p a, bbox_p b)
{
 r->min[0] = vtmin(a->min[0], b->min[0]);
 r->min[1] = vtmin(a->min[1], b->min[1]);
 r->min[2] = vtmin(a->min[2], b->min[2]);
 r->max[0] = vtmax(a->max[0], b->max[0]);
 r->max[1] = vtmax(a->max[1], b->max[1]);
 r->max[2] = vtmax(a->max[2], b->max[2]);
}
*/

inline void bbox_combine(bbox_p r, bbox_p a, bbox_p b)
{
 r->bounds[0][0] = vtmin(a->bounds[0][0], b->bounds[0][0]);
 r->bounds[0][1] = vtmin(a->bounds[0][1], b->bounds[0][1]);
 r->bounds[0][2] = vtmin(a->bounds[0][2], b->bounds[0][2]);
 r->bounds[1][0] = vtmax(a->bounds[1][0], b->bounds[1][0]);
 r->bounds[1][1] = vtmax(a->bounds[1][1], b->bounds[1][1]);
 r->bounds[1][2] = vtmax(a->bounds[1][2], b->bounds[1][2]);
}



int numcompares = 0;

/* the array contains pointers to bounding volumes, which contain bounding boxes, and means that this gets pointers to pointers to bounding volumes. */
/* inline static int bvol_cmp_sub(const void *m1, const void *m2, int axis) */
inline int bvol_cmp_sub(const void *m1, const void *m2, int axis) 
{
 bbox_p bb1 = &((*((bvol_p *) m1))->bbox);
 bbox_p bb2 = &((*((bvol_p *) m2))->bbox);
 vectype bb1a = (bb1->bounds[0][axis] + bb1->bounds[1][axis]) / 2.0;
 vectype bb2a = (bb2->bounds[0][axis] + bb2->bounds[1][axis]) / 2.0;
 numcompares++;
 if(bb1a < bb2a) return -1;
 if(bb1a > bb2a) return 1;
 return 0;
}

/*
inline int bvol_cmp_sub(const void *m1, const void *m2, int axis) 
{
 bbox_p bb1 = &((*((bvol_p *) m1))->bbox);
 bbox_p bb2 = &((*((bvol_p *) m2))->bbox);
 vectype bb1a = (bb1->min[axis] + bb1->max[axis]) / 2.0;
 vectype bb2a = (bb2->min[axis] + bb2->max[axis]) / 2.0;
 numcompares++;
 if(bb1a < bb2a) return -1;
 if(bb1a > bb2a) return 1;
 return 0;
}
*/

/* static int bvol_cmp_x(const void *m1, const void *m2) */
inline int bvol_cmp_x(const void *m1, const void *m2) 
{
 return bvol_cmp_sub(m1, m2, 0);
}

/* static int bvol_cmp_y(const void *m1, const void *m2) */
inline int bvol_cmp_y(const void *m1, const void *m2) 
{
 return bvol_cmp_sub(m1, m2, 1);
}

/* static int bvol_cmp_z(const void *m1, const void *m2) */
inline int bvol_cmp_z(const void *m1, const void *m2) 
{
 return bvol_cmp_sub(m1, m2, 2);
}


/* sorts an array of pointers to bounding volumes */
inline void bvol_sort(bvol_p *bvs, int numbvs, int axis)
{
 switch(axis)
    {
     case 0:
         qsort(bvs, numbvs, sizeof(bvol_p), bvol_cmp_x);
         break;
     case 1:
         qsort(bvs, numbvs, sizeof(bvol_p), bvol_cmp_y);
         break;
     case 2:
         qsort(bvs, numbvs, sizeof(bvol_p), bvol_cmp_z);
         break;
     default:
         fprintf(stderr, "D'oh!!\n");
    }
}

inline void bvol_sort_range(bvol_p *bvs, int start, int end, int axis)
{
 bvol_sort(bvs + start, end - start + 1, axis);
}

void bbox_dump(bbox_p b)
{
 fprintf(stderr, "(%f, %f, %f) - (%f, %f, %f)\n", b->bounds[0][0], b->bounds[0][1], b->bounds[0][2], b->bounds[1][0], b->bounds[1][1], b->bounds[1][2]);
}


/*
void bbox_dump(bbox_p b)
{
 fprintf(stderr, "(%f, %f, %f) - (%f, %f, %f)\n", b->min[0], b->min[1], b->min[2], b->max[0], b->max[1], b->max[2]);
}
*/


bvol_p bvol_buildtree_sub(bvol_p *bvs, int start, int end, int axis)
{
 bvol_p parent;
 if(end < start)
    {
     fprintf(stderr, "Invalid buildtree arguments!\n");
     exit(1);
    }
 if(end == start) return bvs[start];
 parent = bvol_new();
 bvol_sort_range(bvs, start, end, axis);
 axis = (axis + 1) % 3;
 parent->child1 = bvol_buildtree_sub(bvs, start, (start + end) / 2, axis);
 parent->child2 = bvol_buildtree_sub(bvs, (start + end) / 2 + 1, end, axis);
 bbox_combine(&(parent->bbox), &(parent->child1->bbox), &(parent->child2->bbox));
 /* bbox_dump(&(parent->bbox)); */
 return parent;
}


bvol_p bvol_buildtree(bvol_p *bvs, int numbvs)
{
 return bvol_buildtree_sub(bvs, 0, numbvs-1, 0);
}

void bvol_dumptree(bvol_p p, int level)
{
 int t;
 if(p == NULL) return;
 for(t = 0; t < level; t++)
    fprintf(stderr, "  ");
 bbox_dump(&(p->bbox));
 bvol_dumptree(p->child1, level + 1);
 bvol_dumptree(p->child2, level + 1);
}





inline void obj_get_bbox(bbox_p a, obj o)
{
 switch(o->type)
    {
     case obj_type_plane:
         fprintf(stderr, "Can't get bounding box of plane!\n");
         return;
     case obj_type_disc:
         obj_disc_bbox(a, (obj_disc)(o->data));
         return;
     case obj_type_yes:
         return;
     case obj_type_sphere:
         obj_sphere_bbox(a, (obj_sphere)(o->data));
         return;
     case obj_type_triangle:
         obj_triangle_bbox(a, (obj_triangle)(o->data));
         return;
     default:
         fprintf(stderr, "ERROR!  Trying to get bbox for unknown object type!\n");
         return;
    }
}


