/* color operations */

#include "vector.h"
#include "color.h"



color color_new(int type)
{
 color c = malloc(sizeof(struct color_s));
 color_init(c, type);
 return c;
}

void color_free(color c)
{
 c->type = 0;
 free(c);
}


void color_init(color c, int type)
{
 c->type = type;
 color_clear(c);
}


void color_clear(color c)
{
 switch(c->type)
    {
     case COLOR_TYPE_RGB:
         vector_zerow(c->rgba);
         break;
     default:
         fprintf(stderr, "Invalid color type!\n");
    }
}


/* this is BROKEN if you use the alpha channel for transparency.  As it's not
currently used, it's not a problem, but if it is ever used, this will need
major re-working. */
void color_add(color r, color a, color b)
{
 if(r->type != a->type || r->type != b->type)
    {
     fprintf(stderr, "Can't add mismatched color types!\n");
    }
 switch(r->type)
    {
     case COLOR_TYPE_RGB:
         r->rgba[0] = a->rgba[0] * a->rgba[3] + b->rgba[0] * b->rgba[3];
         r->rgba[1] = a->rgba[1] * a->rgba[3] + b->rgba[1] * b->rgba[3];
         r->rgba[2] = a->rgba[2] * a->rgba[3] + b->rgba[2] * b->rgba[3];
         r->rgba[3] = 1.0;
         break;
     default:
         fprintf(stderr, "Invalid color type!\n");
    }
}


void color_scale(color r, color a, vectype b)
{
 if(r->type != a->type)
    {
     fprintf(stderr, "Can't scale mismatched color types!\n");
    }
 switch(r->type)
    {
     case COLOR_TYPE_RGB:
         r->rgba[0] = a->rgba[0] * b;
         r->rgba[1] = a->rgba[1] * b;
         r->rgba[2] = a->rgba[2] * b;
         r->rgba[3] = a->rgba[3];
         break;
     default:
         fprintf(stderr, "Invalid color type!\n");
    }
}


void color_set_rgba(color c, vectype r, vectype g, vectype b, vectype a)
{
 if(c->type != COLOR_TYPE_RGB)
    {
     fprintf(stderr, "Mismatched color type - not RGBA!\n");
    }
 c->rgba[0] = r;
 c->rgba[1] = g;
 c->rgba[2] = b;
 c->rgba[3] = a;
}


void color_set_rgba_vector(color c, vector rgba)
{
 if(c->type != COLOR_TYPE_RGB)
    {
     fprintf(stderr, "Mismatched color type - not RGBA!\n");
    }
 vector_copy(c->rgba, rgba);
}


color color_new_rgba(vectype r, vectype g, vectype b, vectype a)
{
 color c = color_new(COLOR_TYPE_RGB);
 color_set_rgba(c, r, g, b, a);
 return c;
}
