/*
Disc object
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "vector.h"
#include "surface.h"
#include "object.h"
#include "bvol.h"
#include "disc.h"



obj_disc obj_disc_new()
{
 obj_disc d = malloc(sizeof(obj_disc_s));
 return d;
}

/* NOTE!  This _copies_ the plane structure, as the disc data structure contains the plane directly, rather than using pointers...  if you pass an alloced pointer 
to this, don't forgot to free it afterwards! */
obj_disc obj_disc_set_basic(obj_disc o, plane_p p, vectype d)
{
 o->plane = *p;
 o->radius = d;
 return o;
}

/* same note as above */
obj_disc obj_disc_new_set_basic(plane_p p, vectype d)
{
 return obj_disc_set_basic(obj_disc_new(), p, d);
}


/* test if a ray intersects a disc on a plane */
inline vectype test_disc_sub(vector a, plane_p p, ray_p r, vectype d)
{
 vectype t = plane_ints_ray(a, p, r);
 vectype u;
 if(isinf(t) || t < 0)
    {
     /* fprintf(stderr, "DEBUG: test_disc: does not intersect plane (%f)\n", t); */
     return -1;
    }
 u = vector_dist_3d(a, p->point);
 if(u > d)
    {
     /* fprintf(stderr, "DEBUG: test_disc: does not intersect disc (%f > %f)\n", u, d); */
     return -1;
    }
 /* fprintf(stderr, "DEBUG: test_disc: intersects disc %f from center.\n", u); */
 return t;
}

inline vectype test_disc(vector a, obj_disc d, ray_p r)
{
 return test_disc_sub(a, &(d->plane), r, d->radius);
}



inline void obj_disc_bbox(bbox_p a, obj_disc s)
{
 int t;
 for(t = 0; t < 3; t++)
    {
     a->bounds[0][t] = s->plane.point[t] - s->radius * (1.0 - fabs(s->plane.normal[t])) - .000001 * s->radius;
     a->bounds[1][t] = s->plane.point[t] + s->radius * (1.0 - fabs(s->plane.normal[t])) + .000001 * s->radius;
    }
}


/*
inline void obj_disc_bbox(bbox_p a, obj_disc s)
{
 int t;
 for(t = 0; t < 3; t++)
    {
     a->min[t] = s->plane.point[t] - s->radius * (1.0 - fabs(s->plane.normal[t])) - .000001 * s->radius;
     a->max[t] = s->plane.point[t] + s->radius * (1.0 - fabs(s->plane.normal[t])) + .000001 * s->radius;
    }
}
*/


