/*
Objects. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "vector.h"
#include "surface.h"
#include "object.h"

#include "sphere.h"
#include "disc.h"
#include "triangle.h"



obj obj_new()
{
 obj o = malloc(sizeof(obj_s));
 o->type = obj_type_none;
 o->data = NULL;
 o->surf = NULL;
 return o;
}

obj obj_set(obj r, enum objtype t, void *o)
{
 r->type = t;
 r->data = o;
 return r;
}

obj obj_new_set(enum objtype t, void *o)
{
 return obj_set(obj_new(), t, o);
}

obj obj_set_surface(obj o, surface s)
{
 o->surf = s;
 return o;
}



inline vectype obj_test(vector a, obj o, ray_p r)
{
 switch(o->type)
    {
     case obj_type_plane:
         return plane_ints_ray(a, (plane_p)(o->data), r);
     case obj_type_disc:
         return test_disc(a, (obj_disc)(o->data), r);
     case obj_type_yes:
         return 1;
     case obj_type_sphere:
         return test_sphere(a, (obj_sphere)(o->data), r);
     case obj_type_triangle:
         return test_triangle(a, (obj_triangle)(o->data), r);
     default:
         fprintf(stderr, "ERROR!  Test with unknown object type!\n");
         return -1;
    }
}


inline void obj_get_target(vector a, obj o, ray_p r, vectype t)
{
 switch(o->type)
    {
     case obj_type_plane:
     case obj_type_disc:
     case obj_type_triangle:
     case obj_type_yes:
         return;
     case obj_type_sphere:
         ray_extend_3d(a, r, t);
         return;
     default:
         fprintf(stderr, "ERROR!  Getting target point with unknown object type!\n");
         return;
    }
}


/* this would be faster if it returned a pointer rather than copying...  but oh well. */
inline void obj_get_normal(vector a, obj o, vector p)
{
 switch(o->type)
    {
     case obj_type_plane:
         vector_copy_3d(a, ((plane_p)(o->data))->normal);
         return;
     case obj_type_disc:
         vector_copy_3d(a, ((obj_disc)(o->data))->plane.normal);
         return;
     case obj_type_yes:
         return;
     case obj_type_sphere:
         vector_sub_3d(a, p, ((obj_sphere)(o->data))->point);
         vector_norm_3d(a, a);
         return;
     case obj_type_triangle:
         /* vector_copy_3d(a, ((obj_triangle)(o->data))->plane.normal); */
         obj_triangle_get_normal(a, (obj_triangle)(o->data), p);
         return;
     default:
         fprintf(stderr, "ERROR!  Trying to get normal of unknown object type!\n");
         return;
    }
}


