/*
Sphere object
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "vector.h"
#include "surface.h"
#include "object.h"
#include "bvol.h"
#include "sphere.h"


obj_sphere obj_sphere_new()
{
 obj_sphere d = malloc(sizeof(obj_sphere_s));
 return d;
}

obj_sphere obj_sphere_set_basic(obj_sphere o, vector p, vectype r)
{
 vector_copy_3d(o->point, p);
 o->radius = r;
 return o;
}

obj_sphere obj_sphere_new_set_basic(vector p, vectype r)
{
 return obj_sphere_set_basic(obj_sphere_new(), p, r);
}



/* test if a ray intersects a sphere */
inline vectype test_sphere(vector target, obj_sphere sphere, ray_p ray)
{
 vector toc;
 vectype cdist2;
 vectype r2 = sphere->radius * sphere->radius;
 vectype cadist;
 vectype hcdist2;

 vector_sub_3d(toc, sphere->point, ray->point);
 cdist2 = vector_dotprod_3d(toc, toc);
 cadist = vector_dotprod_3d(toc, ray->dir);

 if(cdist2 < r2 + 0.0001)
    {
     /* fprintf(stderr, "DEBUG: test_sphere: originates inside sphere, intersects.\n"); */
     hcdist2 = r2 - cdist2 + cadist*cadist;
     return cadist - sqrt(hcdist2);
    }

 /* fprintf(stderr, "DEBUG: test_sphere: outside sphere.\n"); */
 if(cadist < 0)
    {
     /* fprintf(stderr, "DEBUG: test_sphere: ray points away from sphere, does not intersect.\n"); */
     return -1;
    }

 hcdist2 = r2 - cdist2 + cadist*cadist;
 if(hcdist2 < 0)
    {
     /* fprintf(stderr, "DEBUG: test_sphere: ray misses sphere, does not intersect.\n"); */
     return -1;
    }

 /* fprintf(stderr, "DEBUG: test_sphere: ray intersects sphere!\n"); */
 return cadist - sqrt(hcdist2);
}



inline void obj_sphere_bbox(bbox_p a, obj_sphere s)
{
 a->bounds[0][0] = s->point[0] - s->radius;
 a->bounds[0][1] = s->point[1] - s->radius;
 a->bounds[0][2] = s->point[2] - s->radius;
 a->bounds[1][0] = s->point[0] + s->radius;
 a->bounds[1][1] = s->point[1] + s->radius;
 a->bounds[1][2] = s->point[2] + s->radius;
}



/*
inline void obj_sphere_bbox(bbox_p a, obj_sphere s)
{
 a->min[0] = s->point[0] - s->radius;
 a->min[1] = s->point[1] - s->radius;
 a->min[2] = s->point[2] - s->radius;
 a->max[0] = s->point[0] + s->radius;
 a->max[1] = s->point[1] + s->radius;
 a->max[2] = s->point[2] + s->radius;
}
*/

