/*
Test object generators
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "tga.h"
#include "vector.h"
#include "bvol.h"
#include "world.h"
#include "object.h"
#include "surface.h"
#include "sphere.h"
#include "disc.h"
#include "triangle.h"



void test_sphereflake(world w, vector center, int parentdir, vectype radius, int level)
{
 obj ac;
 surface ae;
 obj_sphere af;
 vector newcenter;
 double factor = 0.5, overlap = 1.0;

 if(level <= 0) return;

 if(1 || level == 1)
    {
     af = obj_sphere_new_set_basic(center, radius);
     if(rand() > RAND_MAX / 2) 
         ae = surface_new_set_rgbdsp(drand() * 0.7 + 0.3, drand() * 0.7 + 0.3, drand() * 0.7 + 0.3, 1.0, 1.0, 100.0);
        else
         ae = surface_new_set_mirror();

     ac = obj_new_set(obj_type_sphere, af);
     obj_set_surface(ac, ae);
     world_add_obj(w, ac);
    }

 /* 0=-z, 1=-x, 2=-y, 3=+z, 4=+x, 5=+y */
 vector_copy_3d(newcenter, center);  newcenter[2] -= radius * (1.0 + factor) - overlap * factor * radius;
 if(parentdir != 0) test_sphereflake(w, newcenter, 3, radius * factor, level - 1);
 vector_copy_3d(newcenter, center);  newcenter[0] -= radius * (1.0 + factor) - overlap * factor * radius;
 if(parentdir != 1) test_sphereflake(w, newcenter, 4, radius * factor, level - 1);
 vector_copy_3d(newcenter, center);  newcenter[1] -= radius * (1.0 + factor) - overlap * factor * radius;
 if(parentdir != 2) test_sphereflake(w, newcenter, 5, radius * factor, level - 1);
 vector_copy_3d(newcenter, center);  newcenter[2] += radius * (1.0 + factor) - overlap * factor * radius;
 if(parentdir != 3) test_sphereflake(w, newcenter, 0, radius * factor, level - 1);
 vector_copy_3d(newcenter, center);  newcenter[0] += radius * (1.0 + factor) - overlap * factor * radius;
 if(parentdir != 4) test_sphereflake(w, newcenter, 1, radius * factor, level - 1);
 vector_copy_3d(newcenter, center);  newcenter[1] += radius * (1.0 + factor) - overlap * factor * radius;
 if(parentdir != 5) test_sphereflake(w, newcenter, 2, radius * factor, level - 1);
}








void test_cubesphere_sub(vector p1, vector p2, vector p3, vector p4, int l, vector hint)
{
 vector p12, p23, p34, p41, pc;
 if(l == 0) fprintf(stdout, "%f,%f,%f %f,%f,%f %f,%f,%f %f,%f,%f\n", p1[0], p1[1], p1[2], p2[0], p2[1], p2[2], p3[0], p3[1], p3[2], p4[0], p4[1], p4[2]);
 if(l == 0) return;

 vector_add_3d(p12, p1, p2);  vector_norm_3d(p12, p12);
 vector_add_3d(p23, p2, p3);  vector_norm_3d(p23, p23);
 vector_add_3d(p34, p3, p4);  vector_norm_3d(p34, p34);
 vector_add_3d(p41, p4, p1);  vector_norm_3d(p41, p41);
 if(hint != NULL)
     vector_norm_3d(pc, hint);
    else
     vector_add_3d(pc, p12, p34); vector_norm_3d(pc, pc);

 test_cubesphere_sub(p1, p12, pc, p41, l-1, NULL);
 test_cubesphere_sub(p2, p23, pc, p12, l-1, NULL);
 test_cubesphere_sub(p3, p34, pc, p23, l-1, NULL);
 test_cubesphere_sub(p4, p41, pc, p34, l-1, NULL);

}

void test_cubesphere()
{
 int l = 6;
 vectype uv = 1.0 / sqrt(3.0);
 vector p1={-uv,-uv,uv}, p2={-uv,uv,uv}, p3={uv,uv,uv}, p4={uv,-uv,uv};
 vector p5={-uv,-uv,-uv}, p6={-uv,uv,-uv}, p7={uv,uv,-uv}, p8={uv,-uv,-uv};

 test_cubesphere_sub(p1, p2, p3, p4, l, NULL);
 test_cubesphere_sub(p5, p6, p7, p8, l, NULL);
 test_cubesphere_sub(p1, p2, p6, p5, l, NULL);
 test_cubesphere_sub(p3, p4, p8, p7, l, NULL);
 test_cubesphere_sub(p1, p4, p8, p5, l, NULL);
 test_cubesphere_sub(p2, p3, p7, p6, l, NULL);

/*
 vectype uv2 = 1.0 / sqrt(2.0);
 vector q1={-uv2,-uv2,0}, q2={-uv2,uv2,0}, q3={uv2,uv2,0}, q4={uv2,-uv2,0};
 vector r1={0,0,-1}, r2={0,0,1};
 test_cubesphere_sub(q1, q2, q3, q4, l, r1); 
 test_cubesphere_sub(q1, q2, q3, q4, l, r2);
*/
}




void test_trisphere_sub(vector p1, vector p2, vector p3, int l)
{
 vector pc;
 if(l == 0) fprintf(stdout, "%f,%f,%f %f,%f,%f %f,%f,%f %f,%f,%f\n", p1[0], p1[1], p1[2], p2[0], p2[1], p2[2], p3[0], p3[1], p3[2], p3[0], p3[1], p3[2]);
 if(l == 0) return;

 vector_add_3d(pc, p1, p2);  
 vector_add_3d(pc, pc, p3);  
 vector_norm_3d(pc, pc);

 test_trisphere_sub(p1, p2, pc, l-1);
 test_trisphere_sub(p2, p3, pc, l-1);
 test_trisphere_sub(p3, p1, pc, l-1);
}

void test_trisphere()
{
 int l = 7;
 vectype uv2 = 1.0 / sqrt(2.0);
 vector q1={-uv2,-uv2,0}, q2={-uv2,uv2,0}, q3={uv2,uv2,0}, q4={uv2,-uv2,0};
 vector r1={0,0,-1}, r2={0,0,1};

 test_trisphere_sub(q1, q2, r1, l);
 test_trisphere_sub(q2, q3, r1, l);
 test_trisphere_sub(q3, q4, r1, l);
 test_trisphere_sub(q4, q1, r1, l);
 test_trisphere_sub(q1, q2, r2, l);
 test_trisphere_sub(q2, q3, r2, l);
 test_trisphere_sub(q3, q4, r2, l);
 test_trisphere_sub(q4, q1, r2, l);
}



void test_trisphere2_sub(vector p1, vector p2, vector p3, int l)
{
 vector p12, p23, p31;
 if(l == 0) fprintf(stdout, "%f,%f,%f %f,%f,%f %f,%f,%f %f,%f,%f\n", p1[0], p1[1], p1[2], p2[0], p2[1], p2[2], p3[0], p3[1], p3[2], p3[0], p3[1], p3[2]);
 if(l == 0) return;

 vector_add_3d(p12, p1, p2);  vector_norm_3d(p12, p12);
 vector_add_3d(p23, p2, p3);  vector_norm_3d(p23, p23);
 vector_add_3d(p31, p3, p1);  vector_norm_3d(p31, p31);

 test_trisphere2_sub(p1, p12, p31, l-1);
 test_trisphere2_sub(p2, p23, p12, l-1);
 test_trisphere2_sub(p3, p31, p23, l-1);
 test_trisphere2_sub(p12, p23, p31, l-1);
}

void test_trisphere2()
{
 int l = 5;
 /* vectype uv2 = 1.0 / sqrt(2.0); */
 /* vector q1={-uv2,-uv2,0}, q2={-uv2,uv2,0}, q3={uv2,uv2,0}, q4={uv2,-uv2,0}; */
 vector q1={-1,0,0}, q2={0,-1,0}, q3={1,0,0}, q4={0,1,0};
 vector r1={0,0,-1}, r2={0,0,1};

 test_trisphere2_sub(q1, q2, r1, l);
 test_trisphere2_sub(q2, q3, r1, l);
 test_trisphere2_sub(q3, q4, r1, l);
 test_trisphere2_sub(q4, q1, r1, l);
 test_trisphere2_sub(q1, q2, r2, l);
 test_trisphere2_sub(q2, q3, r2, l);
 test_trisphere2_sub(q3, q4, r2, l);
 test_trisphere2_sub(q4, q1, r2, l);
}


void test_trisphere3()
{
 int l = 0;
 int a, b, c, t, u;
 vectype gm = (1.0 + sqrt(5)) / 2.0; 
 vector v, v2, v3, v4, v5;

 for(a = 0; a < 3; a++)
    {
     b = (a + 1) % 3;
     c = (a + 2) % 3;
     for(t=0; t<=1; t++)
        {
         for(u=0; u<=1; u++)
            {
             v[a] = t?gm:-gm;
             v[b] = u?1.0:-1.0;
             v[c] = 0.0;
             v3[a] = v2[a] = 0;
             v3[b] = v2[b] = u?gm:-gm;
             v3[c] = 1; v2[c] = -1;
             v5[a] = v4[a] = t?1.0:-1.0;
             v5[b] = v4[b] = 0;
             v5[c] = gm ; v4[c] = -gm;

             vector_norm_3d(v, v);
             vector_norm_3d(v2, v2);
             vector_norm_3d(v3, v3);
             vector_norm_3d(v4, v4);
             vector_norm_3d(v5, v5);

             /*
             fprintf(stderr, "%d %d %d %d %d\n", a, t, u, b, c);
             fprintf(stderr, "%f %f %f - %f %f %f - %f %f %f \n", v[0], v[1], v[2], v2[0], v2[1], v2[2], v3[0], v3[1], v3[2]);
             */

             if(1)
                {
                 test_trisphere2_sub(v, v2, v3, l);
                 if(!a)
                    {
                     test_trisphere2_sub(v, v3, v5, l);
                     test_trisphere2_sub(v, v2, v4, l);
                    }
                }
                else
                {
                 test_trisphere_sub(v, v2, v3, l);
                 if(!a)
                    {
                     test_trisphere_sub(v, v3, v5, l);
                     test_trisphere_sub(v, v2, v4, l);
                    }
                }

            }
        }
    }

}


/* not really a test, but it can live here for now */
void full_spectrum(vectype c, int *r, int *g, int *b)
{
 int points[6][3] = {
     {255, 0,   0},
     {255, 255, 0},
     {0,   255, 0},
     {0,   255, 255},
     {0,   0,   255},
     {255, 0,   255}};
 int numpoints = 6;
 vectype t, shade;
 int point, point2;

 t = c * numpoints;
 point = ((int)t) % numpoints;
 point2 = (point + 1) % numpoints;
 shade = t - (vectype)((int)t);

 *r = (1 - shade) * points[point][0] + shade * points[point2][0];
 *g = (1 - shade) * points[point][1] + shade * points[point2][1];
 *b = (1 - shade) * points[point][2] + shade * points[point2][2];
}



void test_random_discs(world w, int n)
{
 int x;
 vector i, j;
 plane_s p;
 obj_disc ad;
 obj ac;
 surface ae;
 for(x = 0; x < n; x++)
    {
     i[0] = drand() * 10 - 5;
     i[1] = drand() * 10 - 5;
     i[2] = drand() * 10 - 5;
     j[0] = drand() * 2 - 1;
     j[1] = drand() * 2 - 1;
     j[2] = drand() * 2 - 1;
     plane_set(&p, i, j);
     ad = obj_disc_new_set_basic(&p, drand() * 5 + 1);
     ae = surface_new_set_rgbdsp(drand(), drand(), drand(), drand(), drand(), drand()*100);
     ac = obj_new_set(obj_type_disc, ad);
     obj_set_surface(ac, ae);
     world_add_obj(w, ac);
    }
}

void test_random_spheres(world w, int n, int s)
{
 int x;
 vector i;
 obj_sphere af;
 obj ac;
 surface ae;
 for(x = 0; x < n; x++)
    {
     i[0] = drand() * 30 - 15;
     i[1] = drand() * 30 - 17;
     i[2] = drand() * 30 - 15;
     af = obj_sphere_new_set_basic(i, drand() * 1.5 + 0.5);
     if(s)
         ae = surface_new_set_rgbdsp(drand(), drand(), drand(), drand(), drand(), drand()*100);
        else
         ae = surface_new_set_mirror();
     ac = obj_new_set(obj_type_sphere, af);
     obj_set_surface(ac, ae);
     world_add_obj(w, ac);
    }
}

/*
test_trisphere3();
exit(0);
test_trisphere2();
exit(0);
test_trisphere();
exit(0);
test_cubesphere();
exit(0);
*/


void test_random_triangles(world w, int n, int s)
{
 int x;
 vector i, j, k;
 obj_triangle ag;
 obj ac;
 surface ae;
 for(x = 0; x < 0; x++)
    {
     int s = 10;
     i[0] = drand() * s * 2 - s;     i[1] = drand() * s * 2 - s;     i[2] = drand() * s * 2 - s;
     j[0] = drand() * s * 2 - s;     j[1] = drand() * s * 2 - s;     j[2] = drand() * s * 2 - s;
     k[0] = drand() * s * 2 - s;     k[1] = drand() * s * 2 - s;     k[2] = drand() * s * 2 - s;
/*
     i[0] = -5; i[1] = -5; i[2] = -8;
     j[0] = 8;  j[1] = -2; j[2] = -3;
     k[0] = 6;  k[1] = 6;  k[2] = 7;
     i[0] = -5; i[1] = 0;  i[2] = -5;
     j[0] = -5; j[1] = 0;  j[2] = 5;
     k[0] = 5;  k[1] = 0;  k[2] = 0;
*/
     ag = obj_triangle_new_set_basic_vector(i, j, k);
     ae = surface_new_set_rgbdsp(drand() / 2.0 + 0.5, drand() / 2.0 + 0.5, drand() / 2.0 + 0.5, drand(), drand(), drand()*100);
     ac = obj_new_set(obj_type_triangle, ag);
     obj_set_surface(ac, ae);
     world_add_obj(w, ac);
    }
}



void test_twodiscs(world w)
{
 vector i, j;
 plane_s p;
 obj_disc ad;
 obj ac;
 surface ae;

 i[0] = 0; i[1] = 0; i[2] = -13;
 j[0] = 0; j[1] = 0; j[2] = 1;
 plane_set(&p, i, j);
 ad = obj_disc_new_set_basic(&p, 30);
 ae = surface_new_set_rgbdsp(1.0, 0.5, 1.0, 1.0, 1.0, 10.0);
 ac = obj_new_set(obj_type_disc, ad);
 obj_set_surface(ac, ae);
 world_add_obj(w, ac);

 i[0] = 15; i[1] = 0; i[2] = 0;
 j[0] = -1; j[1] = 0; j[2] = 0;
 plane_set(&p, i, j);
 ad = obj_disc_new_set_basic(&p, 30);
 ae = surface_new_set_rgbdsp(0.0, 0.0, 1.0, 1.0, 1.0, 10.0);
 ac = obj_new_set(obj_type_disc, ad);
 obj_set_surface(ac, ae);
 world_add_obj(w, ac);

 i[0] = 0; i[1] = 20; i[2] = 0;
 j[0] = 0; j[1] = -1; j[2] = 0;
 plane_set(&p, i, j);
 ad = obj_disc_new_set_basic(&p, 30);
 ae = surface_new_set_rgbdsp(0.0, 1.0, 0.0, 1.0, 1.0, 10.0);
 ac = obj_new_set(obj_type_disc, ad);
 obj_set_surface(ac, ae);
 world_add_obj(w, ac);

/*
 i[0] = 0; i[1] = -40; i[2] = 0;
 j[0] = 0; j[1] = 1; j[2] = 0;
 plane_set(&p, i, j);
 ad = obj_disc_new_set_basic(&p, 30);
 ae = surface_new_set_rgbdsp(1.0, 0.0, 0.0, 1.0, 1.0, 10.0);
 ac = obj_new_set(obj_type_disc, ad);
 obj_set_surface(ac, ae);
 world_add_obj(w, ac);
*/


/*
 i[0] = 0; i[1] = 0; i[2] = 13;
 j[0] = 0; j[1] = 0; j[2] = -1;
 plane_set(&p, i, j);
 ad = obj_disc_new_set_basic(&p, 30);
 ae = surface_new_set_color(1.0, 1.0, 0.0);
 ac = obj_new_set(obj_type_disc, ad);
 obj_set_surface(ac, ae);
 world_add_obj(w, ac);
*/
/*
 i[0] = 0; i[1] = 20; i[2] = 0;
 j[0] = 0; j[1] = -1; j[2] = 0;
 plane_set(&p, i, j);
 ad = obj_disc_new_set_basic(&p, 30);
 ae = surface_new_set_mirror();
 ac = obj_new_set(obj_type_disc, ad);
 obj_set_surface(ac, ae);
 world_add_obj(w, ac);
*/
}




