/*****
Quarter-assed (that's like half-assed, but worse) TGA routines.
*****/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "tga.h"


/* simply reads an entire file into ram.  returns a pointer to the
data read, nothing more. */
unsigned char *tga_readfile(char *fn)
{
 unsigned char *buf = NULL;
 int fd;
 if((fd = open(fn, O_RDONLY)) >= 0)
     {
      //struct stat st;
      //fstat(fd, &st);
      //buf = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
      //mmap might not work on the solaris boxes, and I'm in no
      //mood to look up compatibilities.  use simple read loop
      //instead.

      int len = 0, r;
      while(1)
          {
           buf = realloc(buf, len + 1030);
           r = read(fd, buf + len, 1024);
           if(r <= 0) break;
           len += r;
          }

      close(fd);
     }
     else
     {
      fprintf(stderr, "Error opening TGA file '%s'.\n", fn);
     }
 return buf;
}


/* gets a pixel from a structure.  probably should define as a
macro for top speed, but inline shouldn't be too bad.  0=R, 1=G,
2=B.  x,y = 0,0 is top, left corner.  does some klugeyness to 
make arguments corespond to expected positions.  */
inline unsigned char tga_getpixel(tga_image *t, int x, int y, int c)
{
 return t->data[((t->height - y - 1) * t->width + x) * 3 + (2 - c)];
}


/* set a pixel.  same argument syntax as above. */
inline unsigned char tga_setpixel(tga_image *t, int x, int y, int c, unsigned char v)
{
 return (t->data[((t->height - y - 1) * t->width + x) * 3 + (2 - c)] = v);
}



/* reads a tga file.  returns a pointer to a tga_image struct. */
tga_image *tga_read(char *fn)
{
 tga_image *t = malloc(sizeof(tga_image));
 unsigned char *f = tga_readfile(fn);
 t->width = f[12] + 256 * f[13];
 t->height = f[14] + 256 * f[15];
 t->data = malloc(t->width * t->height * 3);
 memcpy(t->data, f + 18, t->width * t->height * 3);
 free(f);
 fprintf(stderr, "TGA file '%s' read: %d by %d pixels.\n", fn, t->width, t->height);
 return t;
}


/* writes out a tga file. */
void tga_write(tga_image *t, char *fn)
{
 int fd;
 if((fd = open(fn, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU)) >= 0)
     {
      unsigned char *f = malloc(20);  // header buffer
      memset(f, 0, 20);  // defaults ok for most fields
      f[12] = t->width % 256;
      f[13] = t->width / 256;
      f[14] = t->height % 256;
      f[15] = t->height / 256;
      f[2] = 2;   // image type
      f[16] = 24; // color depth
      write(fd, f, 18);
      write(fd, t->data, t->width * t->height * 3);
      close(fd);
      fprintf(stderr, "TGA file '%s' written: %d by %d pixels.\n", fn, t->width, t->height);
     }
     else
     {
      fprintf(stderr, "Error writing TGA file '%s'.\n", fn);
     }
}


/* new tga file.  returns a pointer to a tga_image struct. */
tga_image *tga_new(int width, int height)
{
 tga_image *t = malloc(sizeof(tga_image));
 t->width = width;
 t->height = height;
 t->data = malloc(t->width * t->height * 3);
 memset(t->data, 0, t->width * t->height * 3);
 fprintf(stderr, "New TGA file: %d by %d pixels.\n", t->width, t->height);
 return t;
}




/* test code. */
/*
int main(int argc, char **argv)
{
 tga_image *t = tga_read("test.tga");
 int x, y;
 for(y = 0; y < t->height ; y++)
     {
      for(x = 0 ; x < t->width; x++)
          {
           fprintf(stderr, "%3d ", tga_getpixel(t, x, y, 0));
          }
      fprintf(stderr, "\n");
     }
      fprintf(stderr, "\n");
 for(y = 0; y < t->height ; y++)
     {
      for(x = 0 ; x < t->width; x++)
          {
           fprintf(stderr, "%3d ", tga_getpixel(t, x, y, 1));
          }
      fprintf(stderr, "\n");
     }
      fprintf(stderr, "\n");
 for(y = 0; y < t->height ; y++)
     {
      for(x = 0 ; x < t->width; x++)
          {
           fprintf(stderr, "%3d ", tga_getpixel(t, x, y, 2));
      tga_setpixel(t, x, y, 0, 255);
          }
      fprintf(stderr, "\n");
     }

 x = y = 0;
 while(x < t->width && y < t->height)
     {
      tga_setpixel(t, x, y, 2, 255);
      x++; y++;
     }

 tga_write(t, "testout.tga");
 return 0;
}
*/

