/*
Triangle object
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector.h"
#include "surface.h"
#include "object.h"
#include "bvol.h"
#include "triangle.h"
#include "vertex.h"



obj_triangle obj_triangle_new()
{
 obj_triangle d = malloc(sizeof(obj_triangle_s));
 return d;
}


void triangle_compute_normal(vector n, vector t1, vector t2, vector t3)
{
 vector l21={0,0,0}, l23={0,0,0};
 vector_sub_3d(l21, t1, t2);
 vector_sub_3d(l23, t3, t2);
 vector_crossprod_3d(n, l23, l21);
 vector_norm_3d(n, n);
}



void obj_triangle_get_normal_patchwork(vector n, obj_triangle o, vector point)
{
 vectype d1, d2, d3;
 d1 = vector_dist_3d(point, o->v1->point);
 d2 = vector_dist_3d(point, o->v2->point);
 d3 = vector_dist_3d(point, o->v3->point);
 
 if(d1 < d2 && d1 < d3) vector_copy_3d(n, o->v1->normal);
 if(d2 < d1 && d2 < d3) vector_copy_3d(n, o->v2->normal);
 if(d3 < d2 && d3 < d1) vector_copy_3d(n, o->v3->normal);
}


void obj_triangle_get_normal(vector n, obj_triangle o, vector point)
{
 vectype a1, a2, a3, at;
 vector l1, l2, l3, l12, l23, l31, c1, c2, c3, s1, s2, s3;

 if(!( (o->v1->props & VERTEX_PROP_NORMAL) && (o->v2->props & VERTEX_PROP_NORMAL) && (o->v3->props & VERTEX_PROP_NORMAL) ))
    {
     triangle_compute_normal(n, o->v1->point, o->v2->point, o->v3->point);
     return;
    }

 vector_sub_3d(l1, o->v1->point, point);
 vector_sub_3d(l2, o->v2->point, point);
 vector_sub_3d(l3, o->v3->point, point);

 vector_sub_3d(l12, o->v2->point, o->v1->point);
 vector_sub_3d(l23, o->v3->point, o->v2->point);
 vector_sub_3d(l31, o->v1->point, o->v3->point);

 vector_crossprod_3d(c1, l1, l12);
 vector_crossprod_3d(c2, l2, l23);
 vector_crossprod_3d(c3, l3, l31);

 a1 = vector_length_3d(c1);
 a2 = vector_length_3d(c2);
 a3 = vector_length_3d(c3);
 at = a1 + a2 + a3;

/*
 fprintf(stderr, "Vector lengths: %f, %f, %f, total %f\n", a1, a2, a3, at);
 fprintf(stderr, "Norms: (%f,%f,%f) (%f,%f,%f) (%f,%f,%f)\n", o->vn1[0], o->vn1[1], o->vn1[2], o->vn2[0], o->vn2[1], o->vn2[2], o->vn3[0], o->vn3[1], o->vn3[2]);
*/

 vector_scale_3d(s1, o->v1->normal, a2 / at);
 vector_scale_3d(s2, o->v2->normal, a3 / at);
 vector_scale_3d(s3, o->v3->normal, a1 / at);

 vector_add_3d(n, s1, s2);
 vector_add_3d(n, n, s3);
 vector_norm_3d(n, n); 
}


/*
void obj_triangle_get_normal_approx(vector n, obj_triangle o, vector point)
{
 vectype a1, a2, a3, at;
 vector s1, s2, s3;
 a1 = plane_eq_eval(o->n23, point);
 a2 = plane_eq_eval(o->n31, point);
 a3 = plane_eq_eval(o->n12, point);
 at = a1 + a2 + a3;
 vector_scale_3d(s1, o->vn1, a1 / at);
 vector_scale_3d(s2, o->vn2, a2 / at);
 vector_scale_3d(s3, o->vn3, a3 / at);
 vector_add_3d(n, s1, s2);
 vector_add_3d(n, n, s3);
  vector_norm_3d(n, n); 
}
*/

/*
void obj_triangle_get_normal_alsobroken(vector n, obj_triangle o, vector point)
{
 vectype u1, u2, d12, d23, p1, p2, d1p;
 vector l12, l23, l1p, n12, n23, np1, np2;

 vector_sub_3d(l12, o->t2, o->t1);
 vector_sub_3d(l23, o->t3, o->t2);
 vector_sub_3d(l1p, point, o->t1);

 d12 = vector_length_3d(l12);
 d23 = vector_length_3d(l23);
 d1p = vector_length_3d(l1p);

 u1 = vector_dotprod_3d(l12, l1p) / d12;


 u2 = vector_dotprod_3d(l23, l1p) / d23;
 p1 = (u1 / d12);
 p2 = (u2 / d23);

 vector_sub_3d(n12, o->vn2, o->vn1);
 vector_sub_3d(n23, o->vn3, o->vn2);

 vector_scale_3d(np1, n12, p1);
 vector_scale_3d(np2, n23, p2);

 vector_add_3d(n, np1, np2);
 vector_add_3d(n, n, o->vn1);
}
*/ 

/*
void obj_triangle_get_normal_broken(vector n, obj_triangle o, vector point)
{
 vectype d1, d2, d3, dt;
 vector s1, s2, s3;
 d1 = vector_dist_3d(point, o->t1);
 d2 = vector_dist_3d(point, o->t2);
 d3 = vector_dist_3d(point, o->t3);
 dt = d1 + d2 + d3;
 vector_scale_3d(s1, o->vn1, 1 - d1 / dt);
 vector_scale_3d(s2, o->vn2, 1 - d2 / dt);
 vector_scale_3d(s3, o->vn3, 1 - d3 / dt);
 vector_add_3d(n, s1, s2);
 vector_add_3d(n, n, s3);
 vector_norm_3d(n, n);
}
*/



void obj_triangle_make_accel(obj_triangle o, obj_triangle_accel a)
{
 vector l12={0,0,0}, l23={0,0,0}, l31={0,0,0} /*, l13={0,0,0}*/;
 vector u1={0,0,0}, u2={0,0,0}, u3={0,0,0};
 vector n;

/*
 vector_add_3d(o->plane.point, o->t1, o->t2);
 vector_add_3d(o->plane.point, o->plane.point, o->t3);
 vector_scale_3d(o->plane.point, o->plane.point, 1.0 / 3.0);
*/

 vector_copy_3d(a->plane.point, o->v1->point);

 vector_sub_3d(l12, o->v2->point, o->v1->point);
 vector_sub_3d(l23, o->v3->point, o->v2->point);
 vector_sub_3d(l31, o->v1->point, o->v3->point);

/*
 vector_sub_3d(l13, t3, t1);
 vector_crossprod_3d(n, l12, l13);
*/
 triangle_compute_normal(n, o->v1->point, o->v2->point, o->v3->point);


 /* fprintf(stderr, "Computed normal (%f,%f,%f).\n", n[0], n[1], n[2]); */

 plane_set(&(a->plane), a->plane.point, n);

 /*
 fprintf(stderr, "l12 l23 l31 l13 (%f,%f,%f) (%f,%f,%f) (%f,%f,%f) (%f,%f,%f).\n", 
         l12[0],l12[1],l12[2], l23[0],l23[1],l23[2], l31[0],l31[1],l31[2], l13[0],l13[1],l13[2]);
 fprintf(stderr, "Plane (%f,%f,%f) (%f,%f,%f,%f).\n", o->plane.point[0],o->plane.point[1],o->plane.point[2], 
         o->plane.normal[0],o->plane.normal[1],o->plane.normal[2],o->plane.normal[3]);
 */

/*
 vector_crossprod_3d(u1, l12, o->plane.normal);
 vector_crossprod_3d(u2, l23, o->plane.normal);
 vector_crossprod_3d(u3, l31, o->plane.normal);
*/

 vector_crossprod_3d(u1, a->plane.normal, l12);
 vector_crossprod_3d(u2, a->plane.normal, l23);
 vector_crossprod_3d(u3, a->plane.normal, l31);

 plane_eq_compute(a->n12, o->v1->point, u1);
 plane_eq_compute(a->n23, o->v2->point, u2);
 plane_eq_compute(a->n31, o->v3->point, u3);
}

obj_triangle_accel obj_triangle_accel_new()
{
 obj_triangle_accel a = malloc(sizeof(obj_triangle_accel_s));
 return a;
}



obj_triangle obj_triangle_set_basic(obj_triangle o, vertex t1, vertex t2, vertex t3)
{
/* vector n; */

 o->v1 = t1;
 o->v2 = t2;
 o->v3 = t3;

/*
 vector_copy_3d(o->t1, t1);
 vector_copy_3d(o->t2, t2);
 vector_copy_3d(o->t3, t3);
*/
 /* fprintf(stderr, "Adding triangle (%f,%f,%f) (%f,%f,%f) (%f,%f,%f).\n", t1[0],t1[1],t1[2], t2[0],t2[1],t2[2], t3[0],t3[1],t3[2]); */
/*
 triangle_compute_normal(n, o->t1, o->t2, o->t3);
 vector_copy_3d(o->vn1, n);
 vector_copy_3d(o->vn2, n);
 vector_copy_3d(o->vn3, n);
*/

 if(0)
    {
     o->accel = obj_triangle_accel_new();
     obj_triangle_make_accel(o, o->accel);
    }
    else
     o->accel = NULL;
 /* fprintf(stderr, "Sides (%f,%f,%f,%f) (%f,%f,%f,%f) (%f,%f,%f,%f).\n", u1[0],u1[1],u1[2],u1[3], u2[0],u2[1],u2[2],u2[3], u3[0],u3[1],u3[2],u3[3]); */

 return o;
}

/*
void obj_triangle_set_vertnorms(obj_triangle o, vector n1, vector n2, vector n3)
{
 vector_copy_3d(o->vn1, n1);
 vector_copy_3d(o->vn2, n2);
 vector_copy_3d(o->vn3, n3);
}
*/

obj_triangle obj_triangle_new_set_basic(vertex t1, vertex t2, vertex t3)
{
 return obj_triangle_set_basic(obj_triangle_new(), t1, t2, t3);
}

obj_triangle obj_triangle_new_set_basic_vector(vector t1, vector t2, vector t3)
{
 vertex v1 = vertex_new(), v2 = vertex_new(), v3 = vertex_new();
 vector_copy_3d(v1->point, t1);
 vector_copy_3d(v2->point, t2);
 vector_copy_3d(v3->point, t3);
 return obj_triangle_set_basic(obj_triangle_new(), v1, v2, v3);
}



#define TRIEPSILON .00005
inline vectype test_triangle(vector target, obj_triangle o, ray_p ray)
{
 vectype side=0, dist; 
 obj_triangle_accel_s al, *a;
 if(o->accel)
    {
     a = o->accel;
    }
    else
    {
     a = &al;
     obj_triangle_make_accel(o, a);
    }
 /* fprintf(stderr, "Testing triangle with ray (%f,%f,%f) -> (%f,%f,%f).\n", ray->point[0],ray->point[1],ray->point[2], ray->dir[0],ray->dir[1],ray->dir[2]); */
 dist = plane_ints_ray_sided(target, &side, &(a->plane), ray);
 /* fprintf(stderr, "Ray intersects plane at %f on side %f.\n", dist, side); */
 if(dist < 0) return dist;
 /* fprintf(stderr, "Planes evaluate to %f %f %f.\n", plane_eq_eval(o->n12, target), plane_eq_eval(o->n23, target), plane_eq_eval(o->n31, target)); */
 if(plane_eq_eval(a->n12, target) < 0 -TRIEPSILON) return -1;
 if(plane_eq_eval(a->n23, target) < 0 -TRIEPSILON) return -1;
 if(plane_eq_eval(a->n31, target) < 0 -TRIEPSILON) return -1;
 /* fprintf(stderr, "Ray intersects triangle!  Dist %f, side %f.\n", dist, side); */
 return dist;
}


inline void obj_triangle_bbox(bbox_p a, obj_triangle s)
{
 int i;
 for(i = 0; i < 3; i++)
    {
     a->bounds[0][i] = vtmin3(s->v1->point[i], s->v2->point[i], s->v3->point[i]) - TRIEPSILON;
     a->bounds[1][i] = vtmax3(s->v1->point[i], s->v2->point[i], s->v3->point[i]) + TRIEPSILON;
    }
}



/*
inline void obj_triangle_bbox(bbox_p a, obj_triangle s)
{
 int i;
 for(i = 0; i < 3; i++)
    {
     a->min[i] = vtmin3(s->t1[i], s->t2[i], s->t3[i]);
     a->max[i] = vtmax3(s->t1[i], s->t2[i], s->t3[i]) ;
    }
}
*/
