/*
Vector math, rays, and planes
*/

#ifndef _YR_VECTOR_H_
#define _YR_VECTOR_H_


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>


/* allows easy switching between float and double.  Everything should use this type. */
typedef float vectype;


/* vectors are always 4 long, just to make everything simpler.  This
is already a pointer type, so it can be modified by function calls. */
typedef vectype vector[4];


void vector_dump_3d(vector a);
void vector_dump_3d_nnl(vector a);
inline void vector_add_3d(vector r, vector a, vector b);
inline void vector_add(vector r, vector a, vector b);
inline void vector_sub_3d(vector r, vector a, vector b);
inline vectype vector_dotprod_3d(vector a, vector b);
inline vectype vector_dotprod_4d(vector a, vector b);
inline vectype vector_length_3d(vector a);
inline void vector_scale_3d(vector r, vector a, vectype b);
inline void vector_scale(vector r, vector a, vectype b);
inline vectype vector_norm_3d(vector r, vector a);
inline void vector_crossprod_3d(vector r, vector a, vector b);
inline void vector_copy(vector r, vector a);
inline void vector_copy_3d(vector r, vector a);
inline vectype vector_dist_3d(vector a, vector b);
inline void vector_invert_3d(vector r, vector a);


/* a ray is a point and a direction */
typedef struct ray_s
    {
     vector point;
     vector dir;
    } ray_s;
typedef struct ray_s *ray_p;


inline void ray_dump_3d(ray_p a);
inline void ray_extend_3d(vector p, ray_p r, vectype l);



/* a plane is a point and a normal */
typedef struct plane_s
    {
     vector point;
     vector normal;
    } plane_s;
typedef struct plane_s *plane_p;


inline void plane_dump(plane_p a);
inline void plane_set(plane_p r, vector p, vector n);
inline vectype plane_eval(plane_p a, vector p);
inline vectype plane_ints_ray_sided(vector a, vectype *b, plane_p p, ray_p r);
inline vectype plane_ints_ray(vector a, plane_p p, ray_p r);
inline void plane_eq_compute(vector r, vector p, vector n);
inline vectype plane_eq_eval(vector n, vector p);



double drand();
double drandn();
inline vectype vtmax(vectype a, vectype b);
inline vectype vtmin(vectype a, vectype b);
inline vectype vtmax3(vectype a, vectype b, vectype c);
inline vectype vtmin3(vectype a, vectype b, vectype c);


inline void vector_swap(vector a, vector b);
inline void vector_norm_w(vector a, vector b);
void vector_zerow(vector v);


#endif
