/*
Yes, It's Furrywolf's Free Raytracer!
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "tga.h"
#include "vector.h"
#include "matrix.h"
#include "bvol.h"
#include "world.h"
#include "object.h"
#include "surface.h"
#include "sphere.h"
#include "triangle.h"
#include "disc.h"
#include "tests.h"
#include "ints.h"
#include "read_ply.h"
#include "image.h"
#include "color.h"




inline void vector_reflect(vector r, vector i, vector n)
{
 vector_scale_3d(r, n, -2.0 * vector_dotprod_3d(n, i));
 vector_add_3d(r, i, r);
}



/* re-write to take light object! */
void shader_sorta_phong(color c, surface surf, vector in, vector normal, vector light)
{
 color_s ct;
 vectype t;
 vector r;

 color_init(&ct, c->type);
 if(surf->diffcolor)
    {
     t = -vector_dotprod_3d(normal, light);
     if(t > 0)
        {
         color_scale(&ct, surf->diffcolor, t * surf->diff);
         color_add(c, c, &ct);
        }
    }
 if(surf->speccolor)
    {
     vector_reflect(r, in, normal);
     t = -vector_dotprod_3d(r, light);
     if(t > 0)
        {
         color_scale(&ct, surf->diffcolor, surf->spec * pow(t, surf->phong));
         color_add(c, c, &ct);
        }
    }
}


/* this REALLY makes me mad. */
void ray_nudge(ray_p r)
{
 vector t;
 vector_scale_3d(t, r->dir, .00001);
 vector_add_3d(r->point, r->point, t);
}


color trace_ray(vectype *z, world w, ray_p r, int ctype, int level)
{
 color c = color_new(ctype), cr;
 color_s ct;
 ints_s ii, ii2;
 vectype zmin;
 int l;
 ray_s lr;
 vector rr;

 vector l1 = {1, 1, -1};
 vector_norm_3d(l1, l1);

 zmin = bvol_ints_first(&ii, w->tree, r);
 if(z != NULL) *z = zmin;
 if(zmin < 0) return c;

 ints_get_target(&ii);
 ints_get_normal(&ii);

 for(l = 0; l < 1; l++)
    {
     vector_copy_3d(lr.point, ii.target);
     vector_invert_3d(lr.dir, l1);
     ray_nudge(&lr);

     zmin = bvol_ints_first_ignore(&ii2, w->tree, &lr, ii.data); 
     if(zmin < 0)
        {
         color_init(&ct, ctype);
         shader_sorta_phong(&ct, ii.data->surf, r->dir, ii.normal, l1);
         color_add(c, c, &ct);
        }
    }

 if(ii.data->surf->mirror > 0.0 && level >= 0)
    {
     vector_reflect(rr, r->dir, ii.normal);
     vector_copy_3d(lr.point, ii.target);
     vector_norm_3d(lr.dir, rr);
     ray_nudge(&lr);
     cr = trace_ray(NULL, w, &lr, ctype, level-1);
     color_add(c, c, cr);
     color_free(cr);
    }

 return c;
}



vectype color_diff(color a, color b)
{
 vectype d = 0.0;
 if(b->type != a->type)
    {
     fprintf(stderr, "Can't compare mismatched color types!\n");
    }
 switch(a->type)
    {
     case COLOR_TYPE_RGB:
         d += sqrt( (b->rgba[0] - a->rgba[0]) * (b->rgba[0] - a->rgba[0]) );
         d += sqrt( (b->rgba[1] - a->rgba[1]) * (b->rgba[1] - a->rgba[1]) );
         d += sqrt( (b->rgba[2] - a->rgba[2]) * (b->rgba[2] - a->rgba[2]) );
         d /= 3.0;
         break;
     default:
         fprintf(stderr, "Invalid color type!\n");
    }
 return d;
}




int main(void)
{
 int iw = 1000, ih = 1000;

 vector c;
/* vector eye = {0, -35, 0}, center = {0, 0, 0}, up = {0, 0, 1}; */
 vector eye = {0, -35, 0}, center = {0, 0, 0}, up = {0, 0, 1};
 vector l1 = {1, 1, -1};
 ray_s s;
 int x, y;
 world_s w;
 matrix cam, invcam;
 color col2;
 image i = image_new(iw, ih, COLOR_TYPE_RGB);
 vectype **aamap, **zmap;
 int aan;

 world_init(&w);
 srand(time(NULL));

 vector_norm_3d(l1, l1);


/*
 test_sphereflake(&w, sc, 0, 8.0, 6);
*/
 test_twodiscs(&w);

/* read_ply(&w, "../cnc/dragon_recon/dragon_vrip.ply"); */
/* read_ply(&w, "../cnc/dragon_recon/dragon_vrip_res4.ply"); */
/* read_ply(&w, "../cnc/bunny/reconstruction/bun_zipper.ply"); */
/* read_ply(&w, "ply/redbox.ply.bz2"); */
/* read_ply(&w, "ply/golfball.ply.bz2"); */
/* read_ply(&w, "ply/bunny2.ply.bz2"); */
/* read_ply(&w, "ply/gipshand.ply.bz2"); */
/* read_ply(&w, "ply/brain.ply.bz2"); */
/* read_ply(&w, "ply/neptune.ply.bz2"); */
 read_ply(&w, "ply/heptoroid.ply.bz2"); 


 fprintf(stderr, "Building tree.\n");
 w.tree = bvol_buildtree(w.bvs, w.numobjs);
 fprintf(stderr, "Rendering with %d objects.\n", w.numobjs);


 matrix_camera_image(cam, eye, center, up, 20, iw, ih);
 matrix_camera_image_inverse(invcam, eye, center, up, 20, iw, ih);


 aamap = malloc(sizeof(vectype *) * iw);
 zmap = malloc(sizeof(vectype *) * iw);
 for(x = 0; x < iw; x++)
    {
     aamap[x] = malloc(sizeof(vectype) * ih);
     zmap[x] = malloc(sizeof(vectype) * ih);
     for(y = 0; y < ih; y++)
        {
         aamap[x][y] = 0.0;
         zmap[x][y] = 0.0;
        }
    }

 for(y = 0; y < ih; y++)
    {
     for(x = 0; x < iw; x++)
        {
         vector_copy_3d(s.point, eye);
         c[0] = x; c[1] = y; c[2] = 0; c[3] = 1;
         matrix_mulmv(s.dir, invcam, c);
         vector_norm_3d(s.dir, s.dir);
         col2 = trace_ray(&(zmap[x][y]), &w, &s, COLOR_TYPE_RGB, 10);
         color_add(&(i->data[x][y]), &(i->data[x][y]), col2);
         color_free(col2);
        }
     fprintf(stderr, ".");
/* usleep(50000); */
    }
 fprintf(stderr, "\n");


 for(y = 0; y < ih; y++)
     for(x = 0; x < iw; x++)
        {
         int ep = 9;
         int x2, y2, ncol = 0;
         vectype ap, ap2, ap3, avgz = 0.0;
         color_s avgcol;  color_init(&avgcol, COLOR_TYPE_RGB);
         for(y2 = y-2; y2 <= y+2; y2++)
             for(x2 = x-2; x2 <= x+2; x2++)
                {
                 if(y2 < 0 || x2 < 0 || y2 >= ih || x2 >= iw) 
                     continue;
                 if(x2 == x && y2 == y)
                     continue;
                 color_add(&avgcol, &avgcol, &(i->data[x2][y2]));
                 avgz += zmap[x2][y2];
                 ncol++;
                }

         color_scale(&avgcol, &avgcol, 1.0 / ncol);
         avgz /= ncol;
         ap2 = (sqrt( color_diff(&(i->data[x][y]), &avgcol) + 1) - 1) * 10;
         ap3 = (  pow( (zmap[x][y]/avgz+avgz/zmap[x][y]-2.0)*pow(10,ep) + 1.0, pow(10,-ep)) - 1.0) * pow(10,ep-1);
         ap = 1.0 * ap2 + 1.0 * ap3;
         ap *= 2.0;

         for(y2 = y-1; y2 <= y+1; y2++)
             for(x2 = x-1; x2 <= x+1; x2++)
                {
                 if(y2 < 0 || x2 < 0 || y2 >= ih || x2 >= iw) 
                     continue;
                 aamap[x2][y2] += ap;
                }
        }

 aan = 0;
/*
 for(y = 0; y < ih; y++)
     for(x = 0; x < iw; x++)
        aan += (int)aamap[x][y];
 fprintf(stderr, "Anti-aliasing with %d extra rays.\n", aan);
*/

{
 int c[1000], a;
 for(a = 0; a < 1000; a++) c[a] = 0;
 for(y = 0; y < ih; y++) for(x = 0; x < iw; x++) { a = ((int)sqrt(aamap[x][y]+0.5))*((int)sqrt(aamap[x][y]+0.5));  if(a<0 || a>=1000) continue; c[a]++;  aan += a;  }
 for(a = 0; a < 1000; a++) fprintf(stderr, "%d ", c[a]);
 fprintf(stderr, "\nAnti-aliasing with %d extra rays.\n", aan);
}

 aan = 0;
 for(y = 0; y < ih; y++)
    {
     for(x = 0; x < iw; x++)
        {
if(1){
         int a;
         vectype x2, y2, tx, ty, ss;
         a = sqrt(aamap[x][y] + 0.5);
         if(a < 1) continue;
         ss = 1.0 / (a + 1.0);
         for(tx = 0; tx < a; tx++)
             for(ty = 0; ty < a; ty++)
                {
                 x2 = (vectype)x + ss * (tx + 1);
                 y2 = (vectype)y + ss * (ty + 1);
                 vector_copy_3d(s.point, eye);
                 // c[0] = x2 + drandn() * ss * 0.25;  c[1] = y2 + drandn() * ss * 0.25;  c[2] = 0;  c[3] = 1; 
                 c[0] = x2;  c[1] = y2;  c[2] = 0;  c[3] = 1;
                 matrix_mulmv(s.dir, invcam, c);
                 vector_norm_3d(s.dir, s.dir);
                 col2 = trace_ray(NULL, &w, &s, COLOR_TYPE_RGB, 10);
                 color_add(&(i->data[x][y]), &(i->data[x][y]), col2);
                 color_free(col2);
                 aan++;
                }
         color_scale(&(i->data[x][y]), &(i->data[x][y]), 1.0 / (a*a + 1));
}

/*
         int a;
         for(a = 0; a < (int)aamap[x][y]; a++)
            {
             vector_copy_3d(s.point, eye);
             c[0] = ((vectype)x) + drandn() * 0.75;  c[1] = ((vectype)y) + drandn() * 0.75;  c[2] = 0;  c[3] = 1;
             matrix_mulmv(s.dir, invcam, c);
             vector_norm_3d(s.dir, s.dir);
             col2 = trace_ray(NULL, &w, &s, COLOR_TYPE_RGB, 10);
             color_add(&(i->data[x][y]), &(i->data[x][y]), col2);
             color_free(col2);
            }
         color_scale(&(i->data[x][y]), &(i->data[x][y]), 1.0 / (((int)aamap[x][y]) + 1));
*/



if(0)color_set_rgba(&(i->data[x][y]), ((int)sqrt(aamap[x][y] + 0.5)) / 2.0, ((int)sqrt(aamap[x][y] + 0.5)) / 2.0 - 1, ((int)sqrt(aamap[x][y] + 0.5)) / 2.0 - 2, 1.0);



/*
         color_set_rgba(&(i->data[x][y]), 0.2, 0.2, 0.2, 1.0);
         color_scale(&(i->data[x][y]), &(i->data[x][y]), (int)sqrt(aamap[x][y]));
*/


        }
     fprintf(stderr, ".");
/* usleep(500000); */
    }
 fprintf(stderr, "\n");

 fprintf(stderr, "%d anti-aliasing rays cast.\n", aan);

 image_write_tga(i, "foo.tga");
 return 0;
}
